{{
OctoPropDemo.spin - demonsrate running up to 8 Propellers simultaneously.
Copyright (c) 2009 by John Steven Denson (jazzed)
See MIT license terms below. 
}}

{{

Current Features:

1) Code is known to work reliably with 5 Propellers ... one line change would allow up to 8.
2) Host read/write to any Device is implemented at up to 2.3MB/s (256 byte transactions).
3) Up to 137 pins are free for the user on 8 Propellers (17 per Propeller + 1).
4) Propellers boot one at a time from EEPROM (about one bootup per second).

Future enhancements:

1) RPC register/run will be added so the "Host" can ask "Devices" to run methods.
2) Asynchronous notifications on the SCL line could be added for "Device" -> "Host" messaging.
3) Up to 5MB/s data transfers are possible (256 byte transactions).
4) Add code to dispatch Host messages to any of 64 COGs.
5) Transaction byte length could be controlled by assertion of the ATN signal instead of
   having a packet byte set the length. There is room in the Device rd/wr buff loops for this. 

This program supports running up to 8 Propellers simultaneously and provides a communications
channel controlled by the Host to other Propeller Devices. To achieve this a Parallel bus
structure with chained reset is used. A 16 bit data architecture would allow up to
"N" Propellers and many transaction types with bus driver code changes. 

An EEPROM is attached to all Propellers pins 28,29. The EEPROM is used for booting.
After boot, pin 28 (I2C SCL) may be used for async signalling.

The primary communications path between Propellers (data plane) is a parallel 8 bit
data bus which (in the first incarnation) is controlled by an ATN and ACK signal.
This allows for Host/Device interaction with only one host. Another design may allow
for bus mastering where any Propeller can take on the Host role (TBD).

Booting all propellers from EEPROM (or download for the host) will be done as follows:

1. The first Propeller to come out of reset (selected by BOEn tied low) will download
the boot program from EEPROM or download from the user PC.

2. The boot program queries the ATN/ACK bits (the BootID) and assign the Propeller ID (PropID).
  
  If the BootID is $3, the booting Propeller takes the Host role and assumes PropID 0;
  else (BootID <> $3) the Propeller takes Device role and asks* the Host for a PropID.

  * The host Propeller starts an instance of FullDuplexSerial on the ATN/ACK pins (BootID)
  with inverted sense (tx/rx idle produces BootID <> $3) as a "DTE" device. The next
  Propeller to boot will detect BootID <> $3 and take the device role which will configure
  an instance of FullDuplexSerial on the BootID pins as "DCE" (the opposite of DTE) so
  that the device can communicate with the host. After the device serial link is up, it
  will post ASCII "+" to the host. The host is waiting for "+", for a predetermined
  number of devices. When the host receives "+" it sends an enumerated PropID ASCII digit
  (1-9) back to the device. The device receives the digit and stores it as a number.
  To allow more than 9 devices, the PropID being sent needs to change to a non-ASCII number.

3. Once the booting Propeller's role is established, it sets the next reset release pin.

  If the propeller is the host, it will go into the device enumeration loop. Once the devices
  have all been enumerated, the ATN/ACK pins are released from serial connection and go high
  while the host bus driver starts.

  If the propeller is a device, it will start polling ATN/ACK (BootID) for $3 again. Once
  $3 is detected on BootID, the device knows it's safe to start the device bus driver.
  While the device is polling the BootID, the device number is emitted on pins P16,17,18 ...
  if these pins need to be used after boot, they should be released.

4. The next Propeller in the chain boots from EEPROM and starts step 2 above.


Once all propellers are booted, normal operation can begin.


Definitions:

Host - The primary controller in a bus structure also known as the Master.
Device - Secondary devices in a bus structure that are not hosts.
Data Plane - A communications path where data transfer is the primary usage.
Control Plane - A communications path where control signaling is the primary usage.
BootID - in this case the BootID is on the ATN/ACK pins and shifted to make 0, 1, 2, or 3.
PropID - the identifier assigned to a given Propeller. PropID is used for parallel bus data.
}}


' Version 1 - Archive  [Date 2009.09.10  Time 07.36]

con
  _CLKMODE      = XTAL1 + PLL16x
  _XINFREQ      = 5_000_000
  '_XINFREQ      = 6_250_000  

  
obj
  'v  : "SpudVMM"               ' uncomment for spud debugger
  's  : "SpudSerial"            ' uncomment for spud debugger 
  s  : "FullDuplexSerial"       ' comment out for spud debugger
  s2 : "FullDuplexSerial"       ' host/device bootup communications channel
  def: "ParallelDefs"           ' defines for Parallel bus, etc...
  ho : "ParallelHost"           ' Parallel Bus "Host" driver
  dv : "ParallelDevice"         ' Parallel Bus "Device" driver

var     ' interface variables. keep in this order
  long  command
  long  table
  long  destination
  long  address
  long  length
  long  bufferptr
  long  propid

var     ' comm buffer
  byte  buffer[256]

con
  LastID          = 5           ' the number of Propellers to boot
  SerATN          = "+"         ' serial attention character
                                           
pub main | id, cog

  ' start normal pin 31/30 serial port in mode 0
  cog := s.start(def#RXPIN,def#TXPIN,def#MODE,def#BAUD)

  'v.start      ' uncomment for debug
  
  waitcnt(clkfreq/10+cnt)       ' wait for FDX cog to start
  id := getBootID               ' get bootID ... obviously.

  if id == 3                    ' if host Propeller,
    waitcnt(clkfreq+cnt)        ' give PC user a chance to connect
    
  s.str(string($d,"OctoProp BootID "))
  s.hex(id,2)
  s.tx(" ")
  
  if id == 3
    initHost(lastID)            ' ATN/ACK pins not low. We are the Host.
  else
    initDevice                  ' A Host is already driving ATN or ACK low. We are a Device. 
    
  repeat                        ' we may never get here depending on the init code
    waitcnt(clkfreq+cnt)        ' generally a safe, low power thing to do.


pub getBootID | n, t
  return (ina >> def#ATNPIN) & 3

pub bootNext
  outa |= 1 << def#RST
  dira |= 1 << def#RST

pub initDevice | id '(last) | id
{{
Initialize a secondary propeller "Device" (one of LastID selected by main)
}}
  s.str(string($d,"OctoProp Device "))
  s2.start(def#DSTX,def#DSRX,def#DMODE,def#BAUD)        ' device pins reversed from host for comm  
  waitcnt(clkfreq/10+cnt)
  s2.tx(SerATN)                                         ' send Serial ATN to get PropID from Host
  repeat
    id := s2.rx
    if id > 0
      quit
  id := id - "0"                                        ' ID is ascii 1-9
  s.hex(id,2)

  s2.stop                                               ' turn off serial port
  waitcnt(clkfreq/10+cnt)
  dira &= (1 << def#ATNPIN)
  dira &= (1 << def#ACKPIN)
  
  propid := id

  bootNext                                              ' startup the next Propeller

  repeat
    showLED(id)
  while getBootID <> 3                                  ' wait for boot to finish

  dv.start(@command)                                    ' start the device communications driver

  repeat
    waitcnt(clkfreq/10+cnt)                             ' do nothing ... could stop cog 0 ?  

    
pub initHost(last) | verb, pid, ch

  s.str(string($d,"OctoProp Host!",$d,"Waiting for devices: "))

  propid := 0       ' we are the host/master controller

  ' use serial mode 3 where sense of serial is inverted
  ' this means that any new device coming up will detect BootID != 3
  '   
  s2.start(def#DSRX,def#DSTX,def#DMODE,def#BAUD)
  waitcnt(clkfreq/5+cnt)
  s2.rxflush

  bootNext

  repeat pid from 1 to last-1   ' run through list of expected PropID's
    repeat
      ch := s2.rx
      if ch == SerATN           ' wait for "+"
        quit
    s2.rxflush
    s.dec(pid)                  
    s.tx(" ")
    s2.tx(pid+"0")              ' if "+" send next ID

  s2.stop                       ' stop second serial connection
  waitcnt(clkfreq/5+cnt)

  outa |= (1 << def#ATNPIN)     ' set pins to let devices know to start bus monitors
  outa |= (1 << def#ACKPIN)
  waitcnt(clkfreq/5+cnt)        ' wait a little before starting host bus driver
    
  ho.start(@command)            ' start host bus driver
  waitcnt(clkfreq/5+cnt)
  
  s.str(string($d,"Starting Bus Test. "))

  verb := 1
  pid := 1
  ch := 0
  repeat constant(128*10/4)
    repeat pid from 1 to last-1
      buffTest(verb,pid,254)
      ifnot ++ch // 128         ' print dot every N'th test
        s.tx(".")

  s.stop
  waitcnt(clkfreq/10+cnt)       
  
  reboot                        ' For test purposes only now. We want to make sure the test works between reboots
  
    
pub buffTest(chk, pid,len) | n, addr
  addr := $4000
  length  := len
  destination := pid
  bufferptr := @buffer

  ' identity map test
  repeat n from 0 to len
    buffer[n] := n
  wrbuffer(@buffer, len, pid, 0, addr)
  ' repeat n from 0 to len
  '   buffer[n] := 0
  rdbuffer(@buffer, len, pid, 0, addr)

  if chk  
    repeat n from 0 to len-1
      if buffer[n] <> n
        s.str(string($d,"PropDev "))
        s.dec(pid)
        s.str(string(" Bus Test error. Expected "))
        s.hex(n,2)
        s.str(string(" got "))
        s.hex(buffer[n],2)
        repeat
          wrbuffer(@buffer, len, pid, 0, addr)
        return
        
{ ' just print dots if pass at caller
    s.str(string($d,"Propeller Device "))
    s.dec(pid)
    s.str(string(" Buffer Test OK.",$d))
'}

pub wrbuffer(bptr, len, pid, cid, addr)
  destination := cid << 3 | pid
  address := addr
  bufferptr := bptr
  length := len
  ho.wrbuffer(@command)

pub rdbuffer(bptr, len, pid, cid, addr)
  destination := cid << 3 | pid
  address := addr
  bufferptr := bptr
  length := len
  ho.rdbuffer(@command)

pub showLED(id)
{{
Just something to show that all Propellers are running
}}
  dira |= def#LEDMSK << def#LEDPIN
  outa |= id << def#LEDPIN
          
{{
 TERMS OF USE: MIT License                                                           

 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the rights
 to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:

 The above copyright notice and this permission notice shall be included in
 all copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 THE SOFTWARE.
}}